#!/usr/bin/env python
"""
Tests that polyhedral template matching works on the simplest possible system.

Name: ptm.py

Description: Part of the Asap test suite.  Tests the analysis.ptm module

Usage: python ptm.py

Expected result: The output should end with 'ALL TESTS SUCCEEDED'.
"""
from __future__ import print_function

from ase.lattice.cubic import FaceCenteredCubic
from asap3.testtools import ReportTest
from asap3.analysis import PTM
import numpy as np

atoms = FaceCenteredCubic("Cu", size=(7,7,7), pbc=False)
x = atoms.get_positions()
#atoms.set_positions(x + 0.01 * np.random.standard_normal(x.shape))

data = PTM(atoms, rmsd_max=0.1, calculate_strains=True)
print("DONE")
cnt = np.bincount(data['structure'])
print(cnt)
print(data['info'])

ReportTest("Number of FCC atoms", cnt[1], 6*6*6*4, 0)
ReportTest("Number of non-classified atoms", cnt[0], len(atoms) - 6*6*6*4, 0)

data = PTM(atoms, rmsd_max=0.1, calculate_strains=True, quick=True)
cnt = np.bincount(data['structure'])
print(cnt)
print(data['info'])

ReportTest("Number of FCC atoms", cnt[1], 6*6*6*4, 0)
ReportTest("Number of non-classified atoms", cnt[0], len(atoms) - 6*6*6*4, 0)

ReportTest.Summary()



"""A calculator based on the OpenKIM project."""
from __future__ import print_function

from ase.calculators.calculator import Calculator
from asap3 import _asap
from asap3.Internal.ListOfElements import ListOfElements
from asap3.Internal.BuiltinPotentials import get_cell_heights, smallest_pbc_direction
from ase.data import atomic_numbers, chemical_symbols
from ase.utils import basestring
import numpy as np
from copy import copy

OpenKIMinfo = _asap.OpenKIMinfo

kim_file_template = """#
# Autogenerated KIM specification file.
#
# Generated by the Asap module OpenKIMcalculator.
#

KIM_API_Version := 1.6.0

#  UNITS
Unit_length      := A
Unit_energy      := eV
Unit_charge      := e
Unit_temperature := K
Unit_time        := ps

SUPPORTED_ATOM/PARTICLES_TYPES:
# Symbol/name               Type                    code
%(symbols)s

CONVENTIONS:
# Name                      Type
ZeroBasedLists              flag
%(neighboraccess)s
%(neighbormethods)s

MODEL_INPUT:
# Name                      Type         Unit                Shape              Requirements
numberOfParticles           integer      none                []
numberOfSpecies             integer      none                []
particleSpecies             integer      none                [numberOfParticles]
coordinates                 double       length              [numberOfParticles,3]
get_neigh                   method       none                []
neighObject                 pointer      none                []
numberContributingParticles integer      none                []
boxSideLengths              double       length              [3]

MODEL_OUTPUT:
# Name                      Type         Unit                Shape              Requirements
destroy                     method       none                []
compute                     method       none                []
cutoff                      double       length              []
%(energy)s
%(forces)s
%(particleEnergy)s
%(virial)s
%(particleVirial)s
"""

element_template = "%-2s                          spec                    %i\n"
output_template = "%-27s double       %-19s [%s]"
nb_template = "%-27s flag\n"

class OpenKIMcalculator(_asap.OpenKIMcalculator):
    """A calculator interfacing to the OpenKIM models.
    
    Parameters:
    
    name: The long name of the KIM Model.
    
    Optional parameters:
    
    atoms:  If set, set_atoms is called immediately.  Default: no atoms set.
    
    allowed:  List of OpenKIM neighbor list keywords, only these will be considered when 
              matching the model.  Default: ALl neighbor list modes are allowed.
              
    allow_mi_opbc:  If set to False, minimum-image orthogonal periodic boundary condition
                    methods are not considered when mathcing the neighbor lists.  Useful
                    if the cell is expected to become skewed during the simulation.

    access:  Set to "iter" or "loca" to restrict the neigbor list access method to
             iterator mode or locator mode, respectively.  Default: not restricted.
             
    stress:  Set to False to refrain from calculate the global virial, even if the model
             supports it.  Default:  True (calculate global virial / stress).
             
    stresses: As above, but for atomic virials / stresses.
    
    verbose:  Set to True to print additional info during neigborlist matching.
    """
    def __init__(self, name, atoms=None, allowed=None, allow_mi_opbc=True, access=None,
                 stress=True, stresses=True, verbose=False):
        _asap.OpenKIMcalculator.__init__(self)
        self.name = name
        self.atoms = None
        self.allowed = allowed
        self.allow_mi_opbc = allow_mi_opbc
        self.access = access
        self.support_stress = stress
        self.support_stresses = stresses
        self.verbose = verbose
        if atoms is not None:
            self.set_atoms(atoms)
            
    def set_atoms(self, atoms, *args):
        """Set the atoms belonging to this calculator.
        
        The function set_atoms(atoms) is defined in C++, and just calls this
        Python function.  This ensures that this Python function is called
        regardless of whether set_atoms() is called from C++ or Python.
        """
        self.kim_spec = self.get_kim_spec(atoms)
        #print self.kim_spec
        self._initialize(self.kim_spec, self.name)
        # Inform the calculator about which quantities should be allocated and calculated.
        for k,v in self.supported_calculations.items():
            self.please_allocate(k, v)
        # Inform the calculator about the translation from atomic numbers to particle types.
        supported_elements = [atomic_numbers[sym] for sym in self.get_supported_types()]
        self.z_to_typecode = {}
        elements = ListOfElements(atoms)
        for e in elements:
            if e not in supported_elements:
                raise RuntimeError("The OpenKIM model '%s' does not support element Z=%i (%s)."
                                   % (self.name, e, chemical_symbols[e]))
            code = self.get_type_code(chemical_symbols[e])
            if code < 0:
                raise ValueError("Negative KIM type codes not supported - what do they mean?")
            self.z_to_typecode[e] = code
            if self.verbose:
                print("Translation: Z = %i -> id = %i" % (e, code))
        self.set_translation(self.z_to_typecode)
        # Find out which neighborlist type we got.  Check sanity and if ImageAtoms should be used.
        nblistmethod = self.get_NBC_method()
        pbc = atoms.get_pbc()
        cutoff = self.get_cutoff()
        if self.verbose:
            print("Neighbor list method: %s;  cutoff: %.3f A;  pbc: %s" % (nblistmethod, cutoff, pbc))
        if nblistmethod == 'CLUSTER':
            if not np.array_equal(pbc, (False, False, False)):
                raise RuntimeError("OpenKIM chose CLUSTER, but PBC are "+str(pbc))
        elif nblistmethod == 'NEIGH_RVEC_H' or nblistmethod == 'NEIGH_RVEC_F':
            if smallest_pbc_direction(atoms) < 3 * cutoff:
                if self.verbose:
                    print("Disabling minimum image convention for", nblistmethod)
                self._use_imageatoms()
        elif nblistmethod == "NEIGH_PURE_H" or nblistmethod == "NEIGH_PURE_F":
            if self.verbose:
                print("Activating Image atoms for", nblistmethod)
            self._use_imageatoms()
        elif nblistmethod == "MI_OPBC_H" or nblistmethod == "MI_OPBC_F":
            if not np.array_equal(pbc, (True, True, True)):
                raise RuntimeError("OpenKIM chose %s, but PBC are %s" (nblistmethod, str(pbc)))
            if smallest_pbc_direction(atoms) < 2 * cutoff:
                raise RuntimeError("OpenKIM chose %s, but system is too small." % (nblistmethod,))
        _asap.OpenKIMcalculator.set_atoms(self, atoms, *args)
        
    def get_kim_spec(self, atoms):
        """Return the KIM specification string.
        
        Side effects:
        self.supported_calculations (dictionary) will indicate which quantities
        can be calculated by the Model.
        self.z_to_typecode (dictionary) gives the translation from the 
        atomic numbers to the type codes of the model.
        """
        kim_variables = {'symbols': '', 'neighbormethods': ''}
        elements = ListOfElements(atoms)
        info = OpenKIMinfo(self.name)
        supported_elements = [atomic_numbers[sym] for sym in info.get_supported_types()]
        self.z_to_typecode = {}
        for e in elements:
            if e not in supported_elements:
                raise RuntimeError("The OpenKIM model '%s' does not support element Z=%i (%s)."
                                   % (self.name, e, chemical_symbols[e]))
            kim_variables['symbols'] += element_template % (chemical_symbols[e], e)
        test_methods = ('energy', 'particleEnergy', 'forces', 'virial', 'particleVirial')
        self.supported_calculations = {}
        for method in test_methods:
            self.supported_calculations[method] = info.get_API_index(method) > 0
        if not self.support_stress:
            self.supported_calculations['virial'] = False
        if not self.support_stresses:
            self.supported_calculations['particleVirial'] = False
        if self.supported_calculations['energy']:
            kim_variables['energy'] = output_template % ('energy', 'energy', '')
        else:
            kim_variables['energy'] = ''
        if self.supported_calculations['particleEnergy']:
            kim_variables['particleEnergy'] = output_template % ('particleEnergy', 'energy', 'numberOfParticles')
        else:
            kim_variables['particleEnergy'] = ''
        if self.supported_calculations['forces']:
            kim_variables['forces'] = output_template % ('forces', 'force', 'numberOfParticles,3')
        else:
            kim_variables['forces'] = ''
        if self.support_stress and self.supported_calculations['virial']:
            kim_variables['virial'] = output_template % ('virial', 'energy', '6')
        else:
            kim_variables['virial'] = ''
        if self.support_stress and self.supported_calculations['particleVirial']:
            kim_variables['particleVirial'] = output_template % ('particleVirial', 'energy', 'numberOfParticles,6')
        else:
            kim_variables['particleVirial'] = ''
        nb_methods = self.find_neighbor_methods(atoms)
        for nbm in nb_methods:
            kim_variables['neighbormethods'] += nb_template % (nbm,)
        if self.access == None:
            kim_variables['neighboraccess'] = "Neigh_IterAccess            flag\nNeigh_LocaAccess            flag"
        elif self.access.lower() == 'iter':
            if self.verbose:
                print("Using Neigh_IterAccess only")
            kim_variables['neighboraccess'] = "Neigh_IterAccess            flag"
        elif self.access.lower() == 'loca':
            if self.verbose:
                print("Using Neigh_LocaAccess only")
            kim_variables['neighboraccess'] = "Neigh_LocaAccess            flag"
        else:
            raise ValueError("Illegal value for neighborlist access method: " + str(self.access))
        return kim_file_template % kim_variables          
        
    def find_neighbor_methods(self, atoms):
        """Find which neighbor methods are relevant for this atom type
        
        Parameters:
        atoms: 
        The atoms object - its cell and PBC are examined.
        
        allowed (optional, default=None): 
        The list of neighborlist methods being considered, in prioritized order.  
        The default (None) means all methods, in the order NEIGH_RVEC_H, 
        NEIGH_PURE_H, NEIGH_RVEC_F, NEIGH_PURE_F, MI_OPBC_H, MI_OPBC_F, 
        CLUSTER.
        
        allow_mi_opbc (optional, default=True):
        If set to False, the methods MI_OPBC_H/F will not be considered.  Normally,
        they are considered if there are full periodic boundary conditions and the
        unit cell is orthorhombic.  However, the unit cell may change during a
        simulation, in those cases preventing these boundary conditions may be
        a good idea.
        """
        if not self.allowed:
            allowed = ["NEIGH_RVEC_H", "NEIGH_PURE_H", "NEIGH_RVEC_F", 
                       "NEIGH_PURE_F", "MI_OPBC_H", "MI_OPBC_F",
                       "CLUSTER"]
        elif isinstance(self.allowed, basestring):
            allowed = [self.allowed,]
        else:
            allowed = copy(self.allowed)  # Will be modified
        remove = []
        uc = atoms.get_cell()
        diagonal = True
        for i in range(3):
            for j in range(3):
                if i != j and np.abs(uc[i,j]) > 1e-15:
                    diagonal = False  
        if not (self.allow_mi_opbc and atoms.get_pbc().all() and diagonal):
            # Minimum Image Orthogonal Periodic Boundary Conditions
            # are not allowed
            remove.extend(["MI_OPBC_H", "MI_OPBC_F"])
        if  atoms.get_pbc().any():
            # Cluster method is not allowed
            remove.append("CLUSTER")
        for rem in remove:
            if rem in allowed:
                allowed.remove(rem)
        if self.verbose:
            print("Allowed PBC:", allowed)
        return allowed
    
    def get_parameters(self, kind='free', integers=[]):
        """Get the parameters of the model as a dictionary.
        
        The optional parameter 'kind' specifies which kind of parameters you get.
        It can take three values:
        'free' (the default).  You only get the "free" parameters, i.e. those that
            can be changed.
        'fixed': You only get the fixed parameters.
        'all': You get both free and fixed parameters.

        The optional argument 'integers' is a list of parameter names that are
        defined as integers in the parameter file.  This should be detected
        automatically once the KIM API specification is updated.
        
        WARNING: Currently using any other value than 'free' is dangerous, as the
        fixed parameter may be of unexpected types.  Once type-checking of parameters
        appear in the KIM API, accessing the fixed parameters will become safe, and 
        the default will change to 'all'.
        
        NOTE: The prefix PARAM_FREE_ or PARAM_FIXED_ is removed from the parameter 
        names.  If kind=='all', free parameters may shadow fixed parameters of the 
        same name.
        """
        if kind not in ['free', 'fixed', 'all']:
            raise ValueError("get_parameters: kind must be 'free', 'fixed' or 'all'.")
        result = {}
        if kind == 'fixed' or kind == 'all':
            prf = 'PARAM_FIXED_'
            for name in self.get_fixed_parameter_names():
                assert name.startswith(prf)
                shortname = name[len(prf):]
                if shortname in integers or name in integers:
                    value = self._get_parameter(name, is_integer=True)
                else:
                    value = self._get_parameter(name)
                result[shortname] = value
        if kind == 'free' or kind == 'all':
            prf = 'PARAM_FREE_'
            for name in self.get_free_parameter_names():
                assert name.startswith(prf)
                shortname = name[len(prf):]
                if shortname in integers or name in integers:
                    value = self._get_parameter(name, is_integer=True)
                else:
                    value = self._get_parameter(name)
                result[shortname] = value
        return result
    
    def set_parameters(self, **kwargs):
        """Set the free parameters of an OpenKIM model.
        
        Examples (these three are equivalent):
        
        model.set_parameters(cutoff = 3.7)
        
        model.set_parameters(PARAM_FREE_cutoff = 3.7)
        
        pars = {'cutoff': 3.7}
        model.set_parameters(**pars)
        
        Note that prefixing the parameter name with PARAM_FREE_ is optional.
        
        If a bad parameter specification is given, and this function fails,
        the model will be left in an undefined state.  That can be fixed by
        calling this function again, possibly without parameters.
        """
        prefix = 'PARAM_FREE_'
        allowed = self.get_free_parameter_names()
        for key, value in kwargs.items():
            if key.startswith(prefix):
                realkey = key
            else:
                realkey = prefix+key
            if realkey not in allowed:
                raise TypeError("Got an unexpected parameter name '{}'".format(key))
            if isinstance(value, float):
                self._set_parameter_scalar(realkey, value)
            elif isinstance(value, np.ndarray):
                self._set_parameter_array(realkey, value)
            else:
                raise TypeError("Unsupported value type - must be float or numpy array.")
        self._reinit()
    
if __name__ == '__main__':
    from ase.lattice.cubic import FaceCenteredCubic
    atoms = FaceCenteredCubic(size=(10,10,10), symbol='Cu')
    print("Creating calculator")
    pot = OpenKIMcalculator('EMT_Asap_Standard_AlAgAuCuNiPdPt__MO_118428466217_000')
    print("Setting atoms")
    atoms.set_calculator(pot)
    print("Calculating energy")
    print(atoms.get_potential_energy())
    print(atoms.get_forces()[10:])
    print(atoms.get_stress())




"""nanocrystalline.py - set up a nanocrystalline sample."""
from __future__ import print_function

from ase.lattice.cubic import BodyCenteredCubic, FaceCenteredCubic
from ase.visualize import view
from ase.optimize import FIRE
from ase import units
from asap3.md.nptberendsen import Inhomogeneous_NPTBerendsen
from asap3 import NeighborCellLocator
import numpy as np

def random_rot():
    """Returns a random rotation matrix."""
    # The random rotation matrix should most elegantly be produced from three random
    # Euler angles (or similar), with suitable non-uniform distributions to compensate
    # for the "north pole singularity".  However, it is easier (but less elegant) to
    # get a random rotation by chaining a large number of not quite uniformly distributed
    # random rotations.

    g = np.identity(3)  # Initial rotation matrix
    N = 100             # Number of rotations
    for i in range(N):
        phi, theta, psi = np.random.random(3) * 2.0 * np.pi
        # First Euler rotation about z in matrix form
        D = np.array(((np.cos(phi), np.sin(phi), 0.),
                      (-np.sin(phi), np.cos(phi), 0.),
                      (0., 0., 1.)))
        # Second Euler rotation about x:
        C = np.array(((1., 0., 0.),
                      (0., np.cos(theta), np.sin(theta)),
                      (0., -np.sin(theta), np.cos(theta))))
        # Third Euler rotation, 2nd rotation about z:
        B = np.array(((np.cos(psi), np.sin(psi), 0.),
                      (-np.sin(psi), np.cos(psi), 0.),
                      (0., 0., 1.)))
        # Total Euler rotation
        A = np.dot(B, np.dot(C, D))
        # Apply rotation
        g = np.dot(g, A)
    return g
    

def bcc_grains(size, perturbation=0.0):
    "Creates a grain layout based on a perturbed BCC lattice."
    graincenters = BodyCenteredCubic(symbol='H', latticeconstant=1.0, size=size)
    graincenters = graincenters.get_positions() + 0.25
    graincenters /= size
    pert = np.random.standard_normal(graincenters.shape) * perturbation
    graincenters += pert
    rotations = [random_rot() for g in graincenters]
    return graincenters, rotations


class make_nanocrystal:
    """Create a nanocrystalline sample.

    Parameters:
      size:      Size of the system, in Angstrom (either a number or three numbers).
      centers:   Positions of grains in scaled coordinates.
      rotations: Rotation matrices for the grains.
      delta (optional):
                 Grains are cut delta Angstrom before the mathematical boundary
                 between them (default: 0).
      min (optional):
                 If two atoms are closer than this distance, one of them is removed
                 (default: 2.0).

    In addition, either the 'unit' or the 'symbol' parameter must be specified.
      unit (optional):
                 A unit cell for building the crystals (Atoms object).
                 MUST be orthorhombic!
      symbol (optional):
                 If unit is None, then an FCC crystal of this element is used.
      latticeconstant (optional):
                 If symbol is specified, this overrides the default lattice constant.

    Returns:
      A factory that creates a nanocrystalline sample when its run() method is called.
    """
    def __init__(self, size, centers, rotations, delta=0.0, min=2.0,
                 unit=None, symbol=None, latticeconstant=None):
        try:
            assert len(size) == 3
        except TypeError:
            # Not a sequence
            size = (size, size, size)
        self.size = np.array(size)
        assert self.size.shape == (3,)
        self.centers = centers
        self.rotations = rotations
        self.delta = delta
        self.min = min
        assert len(centers) == len(rotations)
        self.centers *= self.size
        if unit is None:
            unit = FaceCenteredCubic(symbol=symbol, size=(1,1,1), pbc=True,
                                     latticeconstant=latticeconstant)
        self.unit = unit

    def run(self):
        "Create the nanocrystalline sample"
        result = None
        grainnumber = 0
        for center, rotation in zip(self.centers, self.rotations):
            assert center.shape == (3,)
            assert rotation.shape == (3,3)
            # Make a chunk of crystal sufficiently large, centered at (0,0,0)
            atoms = self.make_bulk()
            # Rotate it
            pos = atoms.get_positions()
            atoms.set_positions(np.dot(atoms.get_positions(), rotation))
            # Cut a grain out of it
            atoms = self.cut_bulk(atoms, center)
            atoms.set_tags(grainnumber * np.ones(len(atoms), int))
            # Add it to the sample
            if result is None:
                result = atoms
                atoms.set_cell(self.size, scale_atoms=False)
                atoms.set_pbc(True)
            else:
                result.extend(atoms)
        self.remove_close(result)
        return result

    def make_bulk(self):
        "Create a sufficiently large bulk sample"
        minsize = np.sqrt((self.size * self.size).sum())
        uc = self.unit.get_cell().diagonal()
        mult = [int(np.ceil(minsize / x)) * 2 + 1 for x in uc]
        atoms = self.unit.repeat(mult)
        atoms.set_pbc(False)
        center = atoms.get_cell().diagonal() / 2.0
        atoms.set_positions(atoms.get_positions() - center)
        return atoms

    def cut_bulk(self, atoms, center):
        "Center 'atoms' at 'center' and cut of any parts closer to any other center."
        # Replicate grain centers (PBC) into 26 neighboring cells.
        #print "grain at %s:" % (str(center),)
        a = (-1, 0, 1)
        ctr = []
        for i in a:
            for j in a:
                for k in a:
                    ctr.append(self.centers + np.array([i,j,k]) * self.size)
        centers = np.concatenate(ctr)
        # Find directions to other grain centers
        rcent = centers - center
        d2 = (rcent * rcent).sum(axis=1)
        assert d2.shape == (len(centers),)
        order = np.argsort(d2)
        # Vectors and distances to neighboring grains
        rvecs = rcent[order[1:]]
        dists = np.sqrt(d2[order[1:]])
        # Now center these atoms at this grain
        atoms.set_positions(atoms.get_positions() + center)
        # For each other grain we now cut away atoms that are beyond the halfway plane.
        for rvec, dist in zip(rvecs, dists):
            r = atoms.get_positions() - center
            r_projected = np.inner(r, rvec) * (1.0 / dist)
            keep = r_projected < (0.5 * dist - self.delta)
            #print "  Cut toward %s: Keeping %i atoms of %i" % (str(rvec), keep.sum(), len(atoms))
            atoms = atoms[keep]
        print("grain at %s: %i atoms." % (str(center), len(atoms)))
        return atoms
    
    def remove_close(self, atoms):
        "Remove atoms that are too close"
        print("Removing atoms that are too close (d_cut = %.3f Angstrom)" % (self.min,))
        remove = {}
        nblist = NeighborCellLocator(self.min, atoms)
        # Loop over all pairs of too-close atoms
        for i in range(len(atoms)):
            for j in nblist[i]:
                # Pair (i,j) is too close
                if not (i in remove or j in remove):
                    if len(remove) % 2:
                        remove[i] = None
                    else:
                        remove[j] = None
        print("Removing %i atoms out of %i" % (len(remove), len(atoms)))
        del atoms[remove.keys()]

def minimize_energy(atoms, nstep, pressure_interval=10):
    bulkmodulus = 140e9 / 1e5  # Bulk modulus of typical metal (Cu), in bar.
    dyn = FIRE(atoms)
    unstress = Inhomogeneous_NPTBerendsen(atoms, 50*units.fs, 0,
                                          taup=5000*units.fs,
                                          compressibility=1/bulkmodulus)
    dyn.attach(unstress.scale_positions_and_cell, interval=10)
    dyn.run(steps=nstep)
        
if __name__ == "__main__":
    from asap3 import EMT
    from asap3.analysis.localstructure import RestrictedCNA, CoordinationNumbers

    centers, rotations = bcc_grains((2,2,2), 0.01)
    #centers, rotations = bcc_grains((1,1,1), 0.)
    fact = make_nanocrystal(50.0, centers, rotations, symbol='Cu', delta=0.35, min=2.0)
    atoms = fact.run()
    atoms.set_scaled_positions(atoms.get_scaled_positions())
    print("Running CNA")
    RestrictedCNA(atoms)
    coord = CoordinationNumbers(atoms)
    print("Total number of atoms", len(atoms))
    for i in range(coord.min(), coord.max()+1):
        print("CN=%-2i: %i" % (i, (coord==i).sum()))
    view(atoms)
    vol_before = atoms.get_volume()
    print("Energy minimization.")
    atoms.set_calculator(EMT())
    minimize_energy(atoms, 10000)
    vol_after = atoms.get_volume()
    print("Volume change: %.1f -> %.1f  (%.1f%%)" % (vol_before, vol_after, 
                                                    100*(vol_after/vol_before - 1)))
    print("Running CNA again")
    RestrictedCNA(atoms)
    coord = CoordinationNumbers(atoms)
    print("Total number of atoms", len(atoms))
    for i in range(coord.min(), coord.max()+1):
        print("CN=%-2i: %i" % (i, (coord==i).sum()))
    view(atoms)



# AtomsInArrays.py  --  Implements a list of atoms where the data is stored
#                       in multiarrays.

# Copyright (C) 1999 CAMP and Jakob Schiotz <schiotz@fysik.dtu.dk>

"""Implements a list of atoms where the data is stored in multiarrays.

Two versions are implemented, one where all information is stored in
cartesian coordinates, and one where they are stored in scaled space.

In both versions most access is done through __getattr__ to data
specified as arguments to the constructor.  If the object was
constructed with xxxx=array then they can be accessed through
GetXxxx() and SetXxxx().  The exception is positions, momenta and
forces that are accessed through GetCartesianXxxx() and
SetCartesianXxxx().

In Asap version 2.9 and later, this module is only used to create
empty atoms for use in MakeParallelAtoms.
"""

import string
import types
import Numeric
import LinearAlgebra
#import AtomInList
#import DocumentedAttributes
#from Structures.TypesFromAtomicNumbers import TypesFromAtomicNumbers
#from Structures.FakeUnitCell import FakeUnitCell
# Delayed imports:
# Visualization.Avatars.RasMol is imported in AtomsInCmFile.GetPlot()

class ListOfCartesianAtoms:
    """A list of atoms where the storage is in the form of multiarrays.

    The object is created like this:
        ListOfCartesianAtoms(positions=data, momenta=data, .....)
        
    The data is then accessed through the usual access functions, i.e.
    GetCartesianPositions(), GetCartesianMomenta(),
    GetCartesianForces().  Other information can be given when the
    object is initialized, for example stresses can be given as
    stesses=data.  They can then be accessed with GetStresses() and
    changed with SetStresses().
    """

    _specials = {"positions":1, "momenta":1, "forces":1}
    
    def __init__(self, periodicboundaries=None, **kargs):
        # Check that positions are specified, and that they are a 3xN
        # multiarray
        if "positions" not in kargs:
            raise ArgumentError("positions missing")
        try:
            shape = kargs["positions"].shape
        except AttributeError:
            raise TypeError("positions should be a multiarray")
        if shape[1] != 3:
            raise ValueError("positions should be an Nx3 multiarray")

	if periodicboundaries is not None:
	    self.SetPeriodicBoundaries(periodicboundaries)
	
        len = shape[0]
        self._len = len
        self._data = {}
        for arg in kargs.keys():
            if kargs[arg] is not None:
                shape = kargs[arg].shape
                if shape[0] != len:
                    msg = ("%s has shape %t, expected (%d, ?)." %
                           (arg, shape, len))
                    raise ValueError(msg)
                self._data[arg] = kargs[arg]

    # List functions
    def __len__(self):
        return self._len

    def __getitem__(self, n):
        if n < 0 or n >= self._len:
            raise IndexError
        return AtomInList.AtomInList(self, n)

    def __getslice__(self, low, high):
        # Select the data to be used in the slice object
        newdata = {}
        for key in self._data.keys():
            newdata[key] = self._data[key][low:high]
        # Now create a new list of atoms object passing the data to the
        # constructor
        return self.__class__(*(), **newdata)

    def extend(self, other):
	"""Extend the list of atoms by adding atoms from another list.

	The other list must at least contain the same type of data as this
	one.
	"""
	newdata = {}
	for key in self._data.keys():
	    accessfunc = "GetCartesian"+string.capitalize(key)
	    try:
		newdata[key] = getattr(other, accessfunc)()
	    except AttributeError:
		accessfunc = "Get"+string.capitalize(key)
		newdata[key] = getattr(other, accessfunc)()
	for key in self._data.keys():
	    self._data[key] = Numeric.concatenate((self._data[key], newdata[key]))

    def __add__(self, other):
	"""Add two lists of atoms."""
	newdata = {}
	for key in self._data.keys():
	    accessfunc = "GetCartesian"+string.capitalize(key)
	    try:
		otherdata = getattr(other, accessfunc)()
	    except AttributeError:
		accessfunc = "Get"+string.capitalize(key)
		otherdata = getattr(other, accessfunc)()
	    selfdata = self._data[key]
	    newdata[key] = Numeric.concatenate((selfdata, otherdata))
	return self.__class__(*(self.GetPeriodicBoundaries(),), **newdata)

    def __delitem__(self, n):
	"""Delete an atom."""
	if n < 0 or n >= self._len:
	    raise IndexError("list deletion index out of range.")
	for key in self._data.keys():
	    d = self._data[key]
	    self._data[key] = Numeric.concatenate((d[:n], d[n+1:]))
	self._len = self._len - 1
	
    # We should also implement access functions that change/delete the
    # elements in the list

    # Some access functions are not handled by the generic ones

    def GetPeriodicBoundaries(self):
        """Get the periodic boundary conditions."""
        return self._periodicboundaries

    def SetPeriodicBoundaries(self, pbc):
        """Set the periodic boundary conditions."""
        if not isinstance(pbc, tuple):
            raise TypeError("Not a tuple")
        if len(pbc) != 3:
            raise ValueError("Not a tuple of length 3")
        self._periodicboundaries = pbc

    # Visualization
    def GetPlot(self, interactive=0, debug=0):
	"""Get a RasMol avatar plotting the simulation."""
	from Visualization.Avatars.RasMol import RasMol
	return RasMol(self, interactive=interactive, debug=debug)

    # Generic access functions
    def _getdata(self):
        if self._attrname in self._data:
            return self._data[self._attrname]
        else:
            # Attempting to access a missing "attribute"
            raise AttributeError(self._requested)

    def _setdata(self, coordinates):
        a = self._attrname
        if a in self._data:
            if coordinates.shape != self._data[a].shape:
                raise ValueError("wrong shape")
        else:
            # Introducing a new sort of data
            if coordinates.shape[0] != self._len:
                raise ValueError("wrong shape")
        self._data[a] = coordinates

    def _getsingledata(self, n):
        if self._attrname in self._data:
            return self._data[self._attrname][n]
        else:
            # Attempting to access a missing "attribute"
            raise AttributeError(self._requested)

    def _setsingledata(self, n, coordinates):
        if self._attrname in self._data:
            self._data[self._attrname][n] = coordinates
        else:
            # Attempting to access a missing "attribute"
            raise AttributeError(self._requested)

    def _isspecial(self):
        """Checks if access through [GS]etCartesianXXXX is meaningful."""
        if self._attrname == "momentums":
            self._attrname = "momenta"
        if self._attrname not in self._specials:
            # The access function is not allowed to exist
            raise AttributeError(self._requested)
        
    def __getattr__(self, name):
        """Implement the access functions.

        Catches names of the type GetXXXX or GetCartesianXXXX
        or the similar SetXXXX names.
        """
        self._requested = name
        if name[:18] == "GetSingleCartesian":
            self._attrname = string.lower(name[18:])+"s"
            self._isspecial()
            return self._getsingledata
        elif name[:18] == "SetSingleCartesian":
            self._attrname = string.lower(name[18:])+"s"
            self._isspecial()
            return self._setsingledata
        elif name[:9] == "GetSingle":
            self._attrname = string.lower(name[9:])+"s"
            return self._getsingledata
        elif name[:9] == "SetSingle":
            self._attrname = string.lower(name[9:])+"s"
            return self._setsingledata
        if name[:12] == "GetCartesian":
            self._attrname = string.lower(name[12:])
            self._isspecial()
            return self._getdata
        elif name[:12] == "SetCartesian":
            self._attrname = string.lower(name[12:])
            self._isspecial()
            return self._setdata
        elif name[:3] == "Get":
            self._attrname = string.lower(name[3:])
            return self._getdata
        elif name[:3] == "Set":
            self._attrname = string.lower(name[3:])
            return self._setdata
        else:
            raise AttributeError(name)

def _lowerplural(str):
    """Returns the plural of an attribute name converted to lower case."""
    try:
        return string.lower(_pluralmap[str])
    except KeyError:
        return string.lower(str+"s")

Attributes = (
    ("Position", "The position."),
    ("Momentum", "The momentum of the atom.", "Momenta"),
    ("Force", "The forces on the atoms."),
    ("Class", "An integer 'classifying' the atoms.", "Classes"),
    ("Type",
     "An object describing the type of the atoms, e.g. the chemical element."),
    ("AtomicNumber", "The atomic number of the atom."),
    ("Energy", "The energy of the atom", "Energies"),
    ("Mass", "The mass of the atom", "Masses"),
    ("Velocity", "The derivative of the position.", "Velocities"))

def _pl(t):
    l = len(t)
    assert l == 2 or l == 3
    if l == 2:
        return (t[0], t[0]+"s")
    elif l == 3:
        return (t[0], t[2])
    else:
        raise AssertionError("Internal error in _pl, length is not 2 or 3.")

Plurals = map(_pl, Attributes)

del _pl

_pluralmap = {}
for _i in Plurals:
    _pluralmap[_i[0]] = _i[1]
del _i





from __future__ import print_function
from asap3 import *
from ase.lattice.cubic import FaceCenteredCubic
from asap3.testtools import ReportTest
import numpy as np
from OpenKIM_modelname import openkimmodel

#set_verbose(1)
step = 500

if OpenKIMsupported:
    pbc_candidates = ["NEIGH_RVEC_H", 
                      "NEIGH_PURE_H",
                      "NEIGH_RVEC_F", 
                      "NEIGH_PURE_F",
                      "MI_OPBC_H", 
                      "MI_OPBC_F",
                      "CLUSTER"
                      ]
else:
    pbc_candidates = []
    print("OpenKIM support is not compiled into Asap.")

for nbltype in pbc_candidates:
    if nbltype.startswith("MI_OPBC"):
        boundaries = ((1,1,1),)
    elif nbltype == "CLUSTER":
        boundaries = ((0,0,0),)
    else:
        boundaries = ((1,1,1), (0,0,0), (0,0,1))
        
    for pbc in boundaries:
        txt = nbltype + (" PBC=%1i%1i%1i " % pbc)
        # Test that EMT reimported through OpenKIM gives the right results.
        atoms_kim = FaceCenteredCubic(size=(10,10,10), symbol='Cu')
        #atoms_kim = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
        #                    size=(30, 30, 30),
        #                    symbol="Cu")
        natoms = len(atoms_kim)
        atoms_kim.set_pbc(pbc)
        r = atoms_kim.get_positions()
        r.flat[:] += 0.1 * np.sin(np.arange(3*natoms))
        atoms_kim.set_positions(r)
        atoms_emt = atoms_kim.copy()
        kim = OpenKIMcalculator(openkimmodel, allowed=nbltype)
        emt = EMT()
        emt.set_subtractE0(False)
        atoms_kim.set_calculator(kim)
        atoms_emt.set_calculator(emt)
        ek = atoms_kim.get_potential_energy()
        ee = atoms_emt.get_potential_energy()
        ReportTest(txt+"Total energy", ek, ee, 1e-8)
        ek = atoms_kim.get_potential_energies()
        ee = atoms_emt.get_potential_energies()
        for i in range(0, natoms, step):
            ReportTest(txt+"Energy of atom %i" % (i,), ek[i], ee[i], 1e-8)
        fk = atoms_kim.get_forces()
        fe = atoms_emt.get_forces()
        n = 0
        for i in range(0, natoms, step):
            n = (n + 1) % 3
            ReportTest(txt+"Force(%i) of atom %i" % (n, i), fk[i, n], fe[i, n], 1e-8)
        sk = atoms_kim.get_stress()
        se = atoms_emt.get_stress()
        for i in range(6):
            ReportTest(txt+"Stress(%i)" % (i,), sk[i], se[i], 1e-8)
        sk = atoms_kim.get_stresses()
        se = atoms_emt.get_stresses()
        for i in range(0, natoms, step):
            n = (n + 1) % 6
            # Volume per atom is not defined the same way: greater tolerance needed
            ReportTest(txt+"Stress(%i) of atom %i" % (n, i), sk[i, n], se[i, n], 1e-3)

    
ReportTest.Summary()



import os
import time
import numpy as np
import asap3
from asap3.mpi import world
from asap3.io.trajectory import _GetAtoms
try:
    from ase.io.bundletrajectory import BundleTrajectory as _BundleTrajectory
    from ase.io.bundletrajectory import PickleBundleBackend
except ImportError:
    class _BundleTrajectory:
        "This version of ASE does not support BundleTrajectory."
        def __init__(self, *args, **kwargs):
            raise NotImplementedError(self.__doc__)
    class PickleBundleBackend:
        "This version of ASE does not support BundleTrajectory."
        def __init__(self, *args, **kwargs):
            raise NotImplementedError(self.__doc__)

    
class BundleTrajectory(_BundleTrajectory, _GetAtoms):
    """Reads and writes atoms into a .bundle directory.

    The BundleTrajectory is an alternative way of storing
    trajectories, intended for large-scale molecular dynamics
    simulations, where a single flat file becomes unwieldy.  Instead,
    the data is stored in directory, a 'bundle' (the name bundle is
    inspired from bundles in Mac OS, which are really just directories
    the user is supposed to think of as a single file-like unit).

    Parameters:

    filename:
        The name of the directory.  Preferably ending in .bundle.

    mode (optional):
        The file opening mode.  'r' means open for reading, 'w' for
        writing and 'a' for appending.  Default: 'r'.  If opening in
        write mode, and the filename already exists, the old file is
        renamed to .bak (any old .bak file is deleted), except if the
        existing file is empty.

    atoms (optional):
        The atoms that will be written.  Can only be specified in
        write or append mode.  If not specified, the atoms must be
        given as an argument to the .write() method instead.

    backup (optional):
        Use backup=False to disable renaming of an existing file.

    split (optional):
        If set to True or False, determines whether a split file
        format is used instead of the normal one.  In the split
        format, each processor in a parallel simulation writes its own
        files inside the BundleTrajectory, instead of leaving all I/O
        to the master.  In not specified, a split format is used if
        more than one million atoms.  Ignored in serial simulations.

    iolimit (optional):
        Limits the number of MPI tasks performing I/O simultaneously,
        to prevent overloading the NFS server.  Only enforced if the
        number of tasks is somewhat larger than the limit.
        
    """
    def __init__(self, filename, mode='r', atoms=None,
                 backup=True, split=None, iolimit=10):
        if split is None:
            # Decide if subtype should be split based on number of atoms
            split = (atoms is not None) and atoms.get_number_of_atoms() > 1000000
        # Never use subtype split for serial simulations
        if not getattr(atoms, "parallel", False):
            split = False
        # Use the collector object to join all data on master if subtype is normal
        # and the simulation is parallel.
        if not split and atoms is not None and getattr(atoms, "parallel", False):
            atoms = asap3.Collector(atoms)
        if split:
            self.subtype = 'split'
        else:
            self.subtype = 'normal'
        # self.iolimit may be needed if reading a split bundle.
        if world.size < 1.5 * iolimit:
            self.iolimit = None
        else:
            self.iolimit = iolimit
        
        _BundleTrajectory.__init__(self, filename, mode, atoms,
                                   backup=backup)
        if self.subtype == 'split':
            self.set_extra_data('ID')  # So the atoms can be sorted when read.
        
    def _set_defaults(self):
        subtype = self.subtype   # Preserve it
        _BundleTrajectory._set_defaults(self)
        self.subtype = subtype
        self.datatypes['forces'] = False

    def _set_backend(self, backend=None):
        """Set the backed doing the actual I/O."""
        if backend is not None:
            self.backend_name = backend
        if self.backend_name == 'pickle':
            if self.subtype == 'normal':
                # Use the standard ASE backend
                self.backend = PickleBundleBackend(self.master)
            elif self.subtype == 'split':
                self.backend = PickleSplitBundleBackend(self.master,
                                                        self.iolimit)
        else:
            raise NotImplementedError(
                "This version of ASE cannot use BundleTrajectory with backend '%s'"
                % self.backend_name)

    def write(self, atoms=None):
        if self.subtype == 'normal' and atoms is not None and getattr(atoms, "parallel", False):
            atoms = asap3.Collector(atoms)
        _BundleTrajectory.write(self, atoms)

    def _make_bundledir(self, filename):
        """Make the main bundle directory.

        Since all MPI tasks might write to it, all tasks must wait for
        the directory to appear.

        For performance reasons, the first frame directory is created immediately.
        """
        assert not os.path.isdir(filename)
        world.barrier()
        if self.master:
            self.log("Making directory "+filename)
            os.mkdir(filename)
            framedir = os.path.join(filename, "F0")
            self.log("Making directory "+ framedir)    
            os.mkdir(framedir)
            world.barrier()
        else:
            world.barrier()
            i = 0
            while not os.path.isdir(filename):
                time.sleep(1)
                i += 1
            if i > 10:
                self.log("Waiting %d seconds for %s to appear!"
                         % (i, filename))

    def _make_framedir(self, frame):
        """Make subdirectory for the frame.

        For a split bundle, all MPI tasks write to the frame
        directory.  The slaves must therefore wait until it becomes
        available.  To minimize the waiting time, frames are
        pre-created.
        """
        if self.subtype == 'split':
            numdirs = 10
        else:
            numdirs = 1
        if self.master:
            for i in range(frame, frame+numdirs):
                framedir = os.path.join(self.filename, "F"+str(i))
                if not os.path.exists(framedir):
                    self.log("Making directory " + framedir)
                    os.mkdir(framedir)
        framedir = os.path.join(self.filename, "F"+str(frame))
        # Wait for the directory to appear
        world.barrier()
        i = 0
        while not os.path.isdir(framedir):
            time.sleep(1)
            i += 1
        if i > 10:
            self.log("Waiting %d seconds for %s to appear!"
                     % (i, framedir))
        return framedir

    def close(self):
        """Clean up when closing."""
        if self.state == 'write' and self.master:
            i = self.nframes
            while True:
                fname = os.path.join(self.filename, "F" + str(i))
                if not os.path.exists(fname):
                    break
                self.log("Closing, removing empty directory "+fname)
                os.rmdir(fname)
                i += 1
        _BundleTrajectory.close(self)

    def __del__(self):
        if self.state != 'closed':
            self.close()

class PickleSplitBundleBackend(PickleBundleBackend):
    """A special backend for writing split bundles (ASAP only)."""
    def __init__(self, master, iolimit):
        # Store if this backend will actually write anything
        self.writesmall = master
        self.writelarge = True
        self.writenonarray = master
        self.iolimit = iolimit
        if iolimit:
            self.iostart = np.round(
                np.linspace(0, world.size, iolimit+1)).astype(int)
            self.iotag = 413151
        self.lastwritedir = None
            
    def write_small(self, framedir, smalldata):
        "Write small data to be written jointly."
        smalldata['fragments'] = world.size
        PickleBundleBackend.write_small(self, framedir, smalldata)
        
    def write(self, framedir, name, data):
        "Write data to separate file."
        if hasattr(data, "shape"):
            # We need to store which kind of data was written in this frame 
            # so NFS synchronization is possible when closing file.
            if framedir != self.lastwritedir:
                self.lastwritedir = framedir
                self.writenames = []
            self.writenames.append(name)
            # As expected, we are writing a NumPy array
            self.iosync_start()
            name = "%s_%d" % (name, world.rank)
            PickleBundleBackend.write(self, framedir, name, data)
            self.iosync_end()
        elif self.writenonarray:
            # If the data is not a NumPy array, only the master writes.
            PickleBundleBackend.write(self, framedir, name, data)

    def read(self, framedir, name):
        "Read data from separate file."
        self.iosync_start()
        x = PickleBundleBackend.read(self, framedir, name)
        self.iosync_end()
        return x
    
    def iosync_start(self):
        "Prevents too many simultaneous IO tasks from trashing server."
        if self.iolimit and world.rank not in self.iostart:
            # I must wait.
            token = np.zeros(1, int)
            world.receive(token, world.rank-1, self.iotag)

    def iosync_end(self):
        if self.iolimit and world.rank+1 not in self.iostart:
            # Another task is waiting for me.
            token = np.zeros(1, int)
            world.send(token, world.rank+1, self.iotag)
            
    def close(self, log=None):
        """Make sure that all data is available on disk for all MPI tasks."""
        if self.lastwritedir:
            for name in self.writenames:
                for part in range(world.size):
                    fname = os.path.join(self.lastwritedir, "%s_%d.pickle" % (name, part))
                    if not os.path.exists(fname):
                        if log:
                            log.write("Task %i is waiting for '%s' to appear.\n" %
                                      (world.rank, fname))
                        for i in range(20):
                            time.sleep(5)
                            if os.path.exists(fname):
                                break
                        if not os.path.exists(fname) and log:
                            log.write("WARNING: Task %i gave up waiting for '%s'.\n" %
                                      (world.rank, fname))
        self.lastwritedir = None  # Do not repeat the above.
        




from __future__ import print_function
from __future__ import absolute_import
from asap3 import NeighborCellLocator
from ase.atoms import Atoms
from ase.lattice.cubic import FaceCenteredCubic
import numpy as np
import scipy.spatial
from . import plane_utils
from . import quat_utils
import math


def remove_close(atoms, min_dist):
    "Remove atoms that are too close"
    print("Removing atoms that are too close (d_cut = %.3f Angstrom)" % (min_dist,))
    remove = set()
    nblist = NeighborCellLocator(min_dist, atoms)
    # Loop over all pairs of too-close atoms
    for i in range(len(atoms)):
        for j in nblist[i]:
            # Pair (i,j) is too close
            if not (i in remove or j in remove):
                if len(remove) % 2:
                    remove.add(i)
                else:
                    remove.add(j)
    print("Removing %i atoms out of %i" % (len(remove), len(atoms)))
    del atoms[list(remove)]
    return atoms

def calculate_grain_bounding_planes(centres, size):

    a = (-1, 0, 1)
    ctr = [c for c in centres]
    for i in a:
        for j in a:
            for k in a:
                if i == j == k == 0: continue
                ctr += [size * [i, j, k] + c for c in centres]
    ctr = np.array(ctr)

    voronoi = scipy.spatial.Voronoi(ctr, furthest_site=False, incremental=False, qhull_options=None)
    regions = [voronoi.regions[voronoi.point_region[i]] for i in range(len(centres))]

    grain_cells = []
    for i, reg in enumerate(regions):

        cell_vertices = np.array([voronoi.vertices[v] for v in reg])
        hull = scipy.spatial.ConvexHull(cell_vertices)

        bounding_planes = [plane_utils.plane_equation(centres[i], h, cell_vertices) for h in hull.simplices]
        grain_cells += [(cell_vertices, bounding_planes)]

    return grain_cells


def create_atoms(centres, rotations, unit, size, min_dist):
    "Create the nanocrystalline sample"

    grain_cells = calculate_grain_bounding_planes(centres, size)

    ntotal = 0
    result = None
    for grain_num, (centre, rotation) in enumerate(zip(centres, rotations)):
        assert centre.shape == (3,)
        assert rotation.shape == (3,3)

        (bounding_vertices, bounding_planes) = grain_cells[grain_num]

        #find maximum extent in [x, y, z]
        U = rotation
        rot_basis = U / np.linalg.norm(U, axis=0)

        proj_vertices = np.dot(bounding_vertices, rot_basis)

        xmin = np.min(proj_vertices[:,0])
        ymin = np.min(proj_vertices[:,1])
        zmin = np.min(proj_vertices[:,2])

        xmax = np.max(proj_vertices[:,0])
        ymax = np.max(proj_vertices[:,1])
        zmax = np.max(proj_vertices[:,2])

        #calculate number of unit cells in each direction
        cubic_cell = [unit.cell[0][0], unit.cell[1][1], unit.cell[2][2]]

        x0 = int(math.floor(xmin / cubic_cell[0]))
        y0 = int(math.floor(ymin / cubic_cell[1]))
        z0 = int(math.floor(zmin / cubic_cell[2]))

        x1 = int(math.ceil(xmax / cubic_cell[0]))
        y1 = int(math.ceil(ymax / cubic_cell[1]))
        z1 = int(math.ceil(zmax / cubic_cell[2]))

        #create cube of atoms    #todo: outer product
        positions = [np.dot(U, (np.dot((i, j, k), unit.cell) + atom_pos)) for i in range(x0, x1+1) for j in range(y0, y1+1) for k in range(z0, z1+1) for atom_pos in unit.get_positions()]
        positions = np.array(positions)
        print(len(positions))

        #remove excess atoms
        in_cell = np.ones(len(positions)).astype(np.int8)
        for (p, n) in bounding_planes:
            in_cell = np.logical_and(in_cell, np.dot(n, (positions - p).T) > 0)

        positions = positions[in_cell]

        atoms = Atoms(positions = positions, cell=size, numbers = 29 * np.ones(len(positions)))
        atoms.set_tags(grain_num * np.ones(len(atoms), int))
        if result is None:
            result = atoms
            atoms.set_cell(size, scale_atoms=False)
            atoms.set_pbc(True)
        else:
            result.extend(atoms)
        print("grain: %d contains %d atoms" % (grain_num, len(positions)))

    print("Total atoms:", len(result))
    result = remove_close(result, min_dist)
    print("Remaining atoms:", len(result))
    return result

def make_nanocrystal(size, centres, rotations, min_dist=2.0, unit=None, symbol=None, latticeconstant=None):
    """Create a nanocrystalline sample.

    Parameters:
      size:      Size of the system, in Angstrom (either a number or three numbers).
      centres:   Positions of grains in scaled coordinates.
      rotations: Rotation matrices for the grains.

      min_dist (optional):
                 If two atoms are closer than this distance, one of them is removed
                 (default: 2.0).

    In addition, either the 'unit' or the 'symbol' parameter must be specified.
      unit (optional):
                 A unit cell for building the crystals (Atoms object).
                 MUST be orthorhombic!
      symbol (optional):
                 If unit is None, then an FCC crystal of this element is used.
      latticeconstant (optional):
                 If symbol is specified, this overrides the default lattice constant.
    """

    assert type(size) == float or type(size) == list
    if type(size) == float:
        size = np.array([size, size, size])
    elif type(size) == list:
        size = np.array(size)
    assert size.shape == (3,)
    assert len(centres) == len(rotations)
    centres *= size
    if unit is None:
        unit = FaceCenteredCubic(symbol=symbol, size=(1,1,1), pbc=True, latticeconstant=latticeconstant)

    return create_atoms(centres, rotations, unit, size, min_dist)

